-- scope stuff
qualityoflife = qualityoflife or {};
local q = qualityoflife;

q.modules = q.modules or {};
q.modules["QoL_ClassicThreat-TBC"] = q.modules["QoL_ClassicThreat-TBC"] or {};
local c = q.modules["QoL_ClassicThreat-TBC"];
c.name = "QoL_ClassicThreat-TBC";
c.version = 9; -- check for toc
c.minCoreVersion = 5;

local FRAMETYPE_PLAYER = "FRAMETYPE_PLAYER";
local FRAMETYPE_COMPACT_RAID = "FRAMETYPE_COMPACT_RAID";

-- main threat cache: memberGuid -> targetGuid -> threatInfo
local threatCache = {};

-- available target unitIds (depends on accuracy settings)
local availableTargetUnitIds = {
    [1] = "target",
    [2] = "targettarget",
    [3] = "pettarget",
    [4] = "pettargettarget",
    [5] = "targettargettarget",
    [6] = "pettargettargettarget",
    [7] = "targettargettargettarget",
    [8] = "pettargettargettargettarget"
};

local threatStatusColors = {
    [0] = {0.69, 0.69, 0.69},
    [1] = {1, 1, 0.47},
    [2] = {1, 0.6, 0},
    [3] = {1, 0, 0}
};

-- init
local function UpdateAddon()
    if not QOL_OPTIONS[c.name][q.META_LASTVERSION] or QOL_OPTIONS[c.name][q.META_LASTVERSION] < c.version then
        -- updates here

        QOL_OPTIONS[c.name][q.META_LASTVERSION] = c.version;
    end
    return true;
end

local function ShowDropDown(self, level, target, name, userData)
    if c.initFinished and target == "player" and UIDROPDOWNMENU_MENU_LEVEL == 1 then
        -- separator and default values
        local info = UIDropDownMenu_CreateInfo();
        info.notCheckable = true;
        info.notClickable = true;
        info.disabled = true;
        UIDropDownMenu_AddButton(info);

        -- title
        local info = UIDropDownMenu_CreateInfo();
        info.notCheckable = true;
        info.notClickable = true;
        info.hasArrow = false;
        info.disabled = false;
        info.isTitle = true;
        info.text = q:GetText("Classic Threat");
        UIDropDownMenu_AddButton(info);

        -- toggle tank mode
        info.notClickable = false;
        info.notCheckable = false;
        info.disabled = false;
        info.isTitle = false;
        info.checked = c:IsTankModeEnabled();
        info.text = q:GetText("Tank Mode");
        info.func = function() c:ToggleTankMode() end
        UIDropDownMenu_AddButton(info);
    end
end

function c:Init()
    if q.initFinished and not c.initFinished and not c.initFailed then
        QOL_OPTIONS = QOL_OPTIONS or {};
        QOL_OPTIONS[c.name] = QOL_OPTIONS[c.name] or {};

        if q.version >= c.minCoreVersion then
            if UpdateAddon() and c:InitOptions() then

                hooksecurefunc("UnitPopup_ShowMenu", function(...) ShowDropDown(...) end);

                QOL_OPTIONS[c.name]["version"] = c.version;

                c.initFinished = true;
                c:Println(q:GetText("Module loaded."));
                return true;
            end
        else
            c:Println(q:GetText(
                "Error while initialization: QoL Core is outdated (version %s)! Minimum required version: %s",
                q.version, c.minCoreVersion));
            c:Println(q:GetText(
                "Maybe another of your QoL addon comes with an older version. Please use the latest QoL_Core folder or download the latest version from CurseForge."));
        end
        c.initFailed = true;
    end
end

local function GetThreat(testedUnitId, targetUnitId)
    local isTanking, status, threatpct, rawthreatpct, threatvalue, percentOfLead = UnitDetailedThreatSituation(
        testedUnitId, targetUnitId);
    if isTanking then
        -- show percentages above 100 if tanking
        percentOfLead = UnitThreatPercentageOfLead(testedUnitId, targetUnitId);
    end
    return isTanking, status, threatpct, rawthreatpct, threatvalue, percentOfLead;
end

local function RemoveMemberFromCache(guid)
    threatCache[guid] = nil;
end

local function RemoveEnemyFromCache(guid)
    for unitGuid, enemyUnitIds in pairs(threatCache) do
        threatCache[unitGuid][guid] = nil;
    end
end

local function GetUnitThreatSituation(testedUnitId, memberUnitIds)
    local testedUnitGuid = UnitGUID(testedUnitId);
    -- add cache for the tested unitId
    if not threatCache[testedUnitGuid] then
        threatCache[testedUnitGuid] = {};
    end

    -- loop through member unitIds
    for memberUnitId, _ in pairs(memberUnitIds) do
        if UnitExists(memberUnitId) then

            -- add or remove cache entry for this member unitId
            local memberGuid = UnitGUID(memberUnitId);
            if UnitIsDead(memberUnitId) then
                RemoveMemberFromCache(memberGuid);
            else

                -- create set of potential enemy unitIds based on the member unitId and chosen accuracy
                local targetUnitIds, accuracy = {};
                if IsInGroup() then
                    accuracy = QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_PARTY][c.OPT_ACCURACY];
                elseif IsInRaid() then
                    accuracy = QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_RAID_COMPACT][c.OPT_ACCURACY];
                else
                    accuracy = QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_PLAYER][c.OPT_ACCURACY];
                end
                for acc = 1, accuracy do
                    tinsert(targetUnitIds, memberUnitId .. availableTargetUnitIds[acc]);
                end

                -- loop through potential enemy unitIds
                for _, targetUnitId in ipairs(targetUnitIds) do
                    if UnitExists(targetUnitId) and not UnitIsPlayer(targetUnitId) and
                        not UnitIsFriend(testedUnitId, targetUnitId) then
                        local targetGuid = UnitGUID(targetUnitId);

                        -- check if this member-enemy combination has already been tested. if not: go on
                        if UnitIsDead(targetUnitId) then
                            RemoveEnemyFromCache(targetUnitId);
                        elseif testedUnitGuid ~= targetGuid then

                            -- add threat information to main threat cache
                            local isTanking, status, threatpct, rawthreatpct, threatvalue, percentOfLead = GetThreat(
                                testedUnitId, targetUnitId);

                            if rawthreatpct or percentOfLead then
                                local threatInfo = {};
                                threatInfo.targetUnitId = targetUnitId;
                                threatInfo.targetGuid = targetGuid;
                                threatInfo.targetName = UnitName(targetUnitId);
                                threatInfo.isTanking = isTanking;
                                threatInfo.status = status or 0;
                                threatInfo.threatpct = threatpct;
                                threatInfo.rawthreatpct = rawthreatpct;
                                threatInfo.threatvalue = threatvalue;
                                threatInfo.percent = percentOfLead;

                                if percentOfLead and percentOfLead > 0 then
                                    threatInfo.threat = percentOfLead;
                                elseif rawthreatpct == 255 then
                                    threatInfo.threat = threatpct; -- solo workaround
                                else
                                    threatInfo.threat = rawthreatpct;
                                end

                                if threatCache[testedUnitGuid] then
                                    -- entry might have been deleted in the mean time due to player death
                                    threatCache[testedUnitGuid][targetGuid] = threatInfo;
                                end
                            end
                        end
                    end
                end
            end
        end
    end

    -- get highest value from all checked combinations for the tested unitId and return
    -- extra tables for alternating display on PlayerFrame
    local highestStatusThreatInfo, highestThreatPctThreatInfo, highestStatusThreatInfoPlayerFrame,
        highestThreatPctThreatInfoPlayerFrame;
    if threatCache[testedUnitGuid] then
        for targetGuid, threatInfo in pairs(threatCache[testedUnitGuid]) do
            if not highestStatusThreatInfo or threatInfo.status > highestStatusThreatInfo.status then
                highestStatusThreatInfo = threatInfo;
            end
            if not highestThreatPctThreatInfo or threatInfo.threat > highestThreatPctThreatInfo.threat then
                highestThreatPctThreatInfo = threatInfo;
            end

            -- player frame info
            if not (testedUnitId == "player" and UnitGUID("target") == targetGuid) then
                if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_PLAYER][c.OPT_TANK_MODE][q:GetPlayerName()] then
                    -- tank mode
                    if not highestStatusThreatInfoPlayerFrame or threatInfo.status <
                        highestStatusThreatInfoPlayerFrame.status then
                        highestStatusThreatInfoPlayerFrame = threatInfo;
                    end
                    if not highestThreatPctThreatInfoPlayerFrame or threatInfo.threat <
                        highestThreatPctThreatInfoPlayerFrame.threat then
                        highestThreatPctThreatInfoPlayerFrame = threatInfo;
                    end
                else
                    -- normal mode
                    if not highestStatusThreatInfoPlayerFrame or threatInfo.status >
                        highestStatusThreatInfoPlayerFrame.status then
                        highestStatusThreatInfoPlayerFrame = threatInfo;
                    end
                    if not highestThreatPctThreatInfoPlayerFrame or threatInfo.threat >
                        highestThreatPctThreatInfoPlayerFrame.threat then
                        highestThreatPctThreatInfoPlayerFrame = threatInfo;
                    end
                end
            end
        end
    end

    return highestStatusThreatInfo, highestThreatPctThreatInfo, highestStatusThreatInfoPlayerFrame,
        highestThreatPctThreatInfoPlayerFrame;
end

local function ShowTargetThreat()
    if TargetFrame:IsVisible() then

        local frame = c:GetTargetThreatFrame();
        if UnitExists("target") and not UnitIsPlayer("target") and not UnitIsFriend("player", "target") then

            local isTanking, status, threatpct, rawthreatpct, threatvalue, percentOfLead = GetThreat("player", "target");
            status = status or 0;
            rawthreatpct = rawthreatpct or 0;
            if percentOfLead then
                rawthreatpct = percentOfLead;
            end

            if status and rawthreatpct > 0 then
                frame.text:SetText(q:Round(rawthreatpct, 0) .. "%");
                if status > 0 then
                    frame.background:SetVertexColor(unpack(threatStatusColors[status]));
                    frame.background:Show();
                    frame.glow:SetVertexColor(unpack(threatStatusColors[status]));
                    frame.glow:Show();
                else
                    frame.background:Hide();
                    frame.glow:Hide();
                end
                frame:Show();
                return;
            end
        end
        if frame:IsVisible() then
            frame:Hide();
        end
    end
end

local function ShowThreat(unitId, frame, highestStatusThreatInfo, highestThreatPctThreatInfo, frameType)
    if frame then
        if UnitExists(unitId) and not UnitIsDead(unitId) and not UnitIsDead("player") then
            local status, threat = 0, 0;
            if highestStatusThreatInfo then
                status = highestStatusThreatInfo.status or status;
            end
            if highestThreatPctThreatInfo then
                threat = highestThreatPctThreatInfo.threat or threat;
            end

            if status and threat > 0 then
                -- target button (experimental)
                if frame.targetButton then
                    if highestStatusThreatInfo then
                        frame.targetButton:SetAttribute("unit", highestStatusThreatInfo.targetUnitId);
                    elseif highestThreatPctThreatInfo then
                        frame.targetButton:SetAttribute("unit", highestThreatPctThreatInfo.targetUnitId);
                    end
                end

                -- display data
                frame.text:SetText(q:Round(threat, 0) .. "%");

                -- styling
                if status > 0 then
                    if frameType == FRAMETYPE_COMPACT_RAID then
                        frame.text:SetFontObject(nil);
                        if status > 1 then
                            frame.text:SetFontObject("GameFontNormalLarge");
                        else
                            frame.text:SetFontObject("GameFontNormal");
                        end
                        frame.text:SetTextColor(unpack(threatStatusColors[status]));
                        frame.background:Hide();
                    else
                        frame.background:SetVertexColor(unpack(threatStatusColors[status]));
                        frame.background:Show();
                    end
                else
                    frame.text:SetFontObject(nil);
                    if frameType == FRAMETYPE_PLAYER then
                        frame.text:SetFontObject("GameFontNormal");
                    else
                        frame.text:SetFontObject("GameFontNormalSmall");
                    end
                    frame.text:SetTextColor(1, 1, 1);
                    frame.background:Hide();
                end

                -- set outline if enabled
                if frameType == FRAMETYPE_COMPACT_RAID then
                    local fontName, fontHeight, fontFlags = frame.text:GetFont();
                    --frame.text:SetFont(fontName, fontHeight);
                    if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_RAID_COMPACT][c.OPT_OUTLINE] then
                        frame.text:SetFont(fontName, fontHeight, "OUTLINE");
                    end
                end

                frame:Show();
                return;
            end
        end

        frame:Hide();
    end
end

local function UpdateThreat()
    -- target threat
    if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_TARGET][c.OPT_ENABLED] then
        ShowTargetThreat();
    end

    -- create set of member unitIds (self, allies in party or raid and their pets etc.)
    local memberUnitIds = {};
    if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_PLAYER][c.OPT_ENABLED] then
        memberUnitIds["player"] = "player";
    end
    if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_PLAYER_PET][c.OPT_ENABLED] then
        memberUnitIds["playerpet"] = "playerpet";
    end
    if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_TARGET_TARGET][c.OPT_ENABLED] then
        -- TargetFrameToT and TargetFrameToT:IsVisible() and 
        memberUnitIds["targettarget"] = "targettarget";
    end
    if IsInRaid() and QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_RAID_COMPACT][c.OPT_ENABLED] then
        for playerRaidId, playerName in ipairs(q:GetRaidMembers()[0]) do
            memberUnitIds["raid" .. playerRaidId] = playerRaidId;
        end
    end
    if IsInGroup() and QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_PARTY][c.OPT_ENABLED] then
        for playerPartyId, playerName in pairs(q:GetPartyMembers()) do
            memberUnitIds["party" .. playerPartyId] = playerPartyId;
        end
    end
    if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_NAMEPLATE][c.OPT_ENABLED] then
        for unitId, namePlate in pairs(q:GetVisibleNamePlates()) do
            -- print(unitId)
            memberUnitIds[unitId] = namePlate;
        end
    end

    -- get threat info and display
    for memberUnitId, arg in pairs(memberUnitIds) do
        -- print(memberUnitId, arg)
        if UnitExists(memberUnitId) then
            local highestStatusThreatInfo, highestThreatPctThreatInfo, highestStatusThreatInfoPlayerFrame,
                highestThreatPctThreatInfoPlayerFrame = GetUnitThreatSituation(memberUnitId, memberUnitIds);

            if memberUnitId == "player" then
                ShowThreat(memberUnitId, c:GetPlayerThreatFrame(), highestStatusThreatInfoPlayerFrame,
                    highestThreatPctThreatInfoPlayerFrame, FRAMETYPE_PLAYER);
                ShowThreat(memberUnitId, c:GetCompactRaidThreatFrame(arg, memberUnitId), highestStatusThreatInfo,
                    highestThreatPctThreatInfo, FRAMETYPE_COMPACT_RAID);
            elseif memberUnitId == "playerpet" then
                ShowThreat(memberUnitId, c:GetPlayerPetThreatFrame(), highestStatusThreatInfo,
                    highestThreatPctThreatInfo);
            elseif memberUnitId == "targettarget" then
                ShowThreat(memberUnitId, c:GetTargetTargetThreatFrame(), highestStatusThreatInfo,
                    highestThreatPctThreatInfo);
            elseif q:StartsWith(memberUnitId, "party") then
                local frame = c:GetCompactRaidThreatFrame(arg, memberUnitId);
                if frame then
                    ShowThreat(memberUnitId, frame, highestStatusThreatInfo, highestThreatPctThreatInfo,
                        FRAMETYPE_COMPACT_RAID);
                else
                    ShowThreat(memberUnitId, c:GetPartyThreatFrame(arg, memberUnitId), highestStatusThreatInfo,
                        highestThreatPctThreatInfo);
                end
            elseif q:StartsWith(memberUnitId, "raid") then
                ShowThreat(memberUnitId, c:GetCompactRaidThreatFrame(arg, memberUnitId), highestStatusThreatInfo,
                    highestThreatPctThreatInfo, FRAMETYPE_COMPACT_RAID);
            elseif q:StartsWith(memberUnitId, "nameplate") then
                ShowThreat(memberUnitId, c:GetNamePlateThreatFrame(arg), highestStatusThreatInfo,
                    highestThreatPctThreatInfo);
            end
        end
    end
end

local function Reset()
    threatCache = {};
    c:HideFrames();
end

local function OnCombatEvent(event, ...)
    local timestamp, subevent, _, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags,
        destRaidFlags = ...;
    if subevent == "UNIT_DIED" and threatCache then
        -- timer functions are neccessary because async(?) threat methods have noticable delay
        if destGUID == UnitGUID("player") then
            -- player dies
            Reset();
            C_Timer.After(1, function()
                Reset();
            end);
        elseif q:StartsWith(destGUID, "Player") then
            -- might have been a member who died
            if IsInGroup() or IsInRaid() then
                RemoveMemberFromCache(destGUID);
                UpdateThreat();
                C_Timer.After(1, function()
                    RemoveMemberFromCache(destGUID);
                    UpdateThreat();
                end);
            end
        else
            -- enemy died: remove from cache
            RemoveEnemyFromCache(destGUID);
            UpdateThreat();
            C_Timer.After(1, function()
                RemoveEnemyFromCache(destGUID);
                UpdateThreat();
            end);
        end
    end
end

-- command handling
SLASH_CLASSICTHREAT1 = "/classicthreat";
local CMD_TANKMODE = "tankmode";

SlashCmdList["CLASSICTHREAT"] = function(msg)
    if msg then
        if c.initFinished then
            msg = q:Trim(msg);

            if msg == CMD_TANKMODE then
                c:ToggleTankMode();
            else
                c:Println(q:GetText("Unknown command. Possible parameters are: /classicthreat ..."));
                c:Println(q:GetText("tankmode -> Toggles player tank mode."));
            end
        else
            c:Println(q:GetText("The AddOn has not been initialized due to an error. Please relog and try again."));
        end
    end
end

local eventFrame;
local function HookEvents()
    if not eventFrame then
        eventFrame = q:GetEventFrame();
        eventFrame:RegisterEvent("UNIT_THREAT_LIST_UPDATE");
        eventFrame:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE");
        eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
        eventFrame:RegisterEvent("PLAYER_REGEN_ENABLED");
        eventFrame:RegisterEvent("PLAYER_REGEN_DISABLED");
        eventFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
        eventFrame:RegisterEvent("PLAYER_ENTERING_WORLD");
        -- eventFrame:RegisterEvent("NAME_PLATE_UNIT_ADDED");
        -- eventFrame:RegisterEvent("NAME_PLATE_UNIT_REMOVED");
        eventFrame:HookScript("OnEvent", function(self, event, arg1, arg2, arg3, arg4)
            if event then
                if event == "ADDON_LOADED" and arg1 == c.name then
                    c:Init();
                elseif c.initFinished then
                    if event == "PLAYER_ENTERING_WORLD" or event == "PLAYER_REGEN_ENABLED" or event ==
                        "PLAYER_REGEN_DISABLED" then
                        Reset();
                    elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
                        OnCombatEvent(event, CombatLogGetCurrentEventInfo());
                    elseif event == "UNIT_THREAT_LIST_UPDATE" or event == "UNIT_THREAT_SITUATION_UPDATE" then
                        UpdateThreat();
                    elseif event == "PLAYER_TARGET_CHANGED" then
                        if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_PLAYER][c.OPT_ENABLED] then
                            ShowThreat("player", c:GetPlayerThreatFrame());
                        end
                        if QOL_OPTIONS[c.name][c.PROFILE_DEFAULT][c.FRAME_TARGET][c.OPT_ENABLED] then
                            ShowTargetThreat();
                        end
                    end
                end
            end
        end);
    end
end
HookEvents();
