qualityoflife = qualityoflife or {};
local q = qualityoflife;

local oldPartyMembers, oldRaidMembers;
local lastJoinedMember, lastLeftMember;

function q:GetPlayerName()
	return q:GetCharName();
end

function q:GetCharName()
	if not q.playerName then
		q.playerName = GetUnitName("player");
	end
	return q.playerName;
end

function q:GetPartyMembers()
	local partyMembers = {};
	for i = 1, 4 do
		local playerName = UnitName('party' .. i);
		if playerName then
			partyMembers[i] = playerName;
		end
	end
	return partyMembers;
end

function q:GetRaidMembers()
	local raidMembers = { [0] = {} };
	local noRaidMembers = GetNumGroupMembers();

	if noRaidMembers then
		for n = 1, noRaidMembers do
			local name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(n);

			-- table of all raid members by id
			raidMembers[0][n] = name;

			-- table of all raid members, mapped by subgroups
			if not raidMembers[subgroup] then
				raidMembers[subgroup] = {};
			end
			local subGroupId = table.getn(raidMembers[subgroup]) + 1;
			raidMembers[subgroup][subGroupId] = name;

			n = n + 1;
		end
	end
	return raidMembers;
end

function q:ResetMemberInfo()
	oldPartyMembers, oldRaidMembers = q:Deepcopy(q:GetPartyMembers()), q:Deepcopy(q:GetRaidMembers());
	--partyMembers, raidMembers = nil, nil;
	lastJoinedMember, lastLeftMember = nil, nil;
end

function q:GetLastJoinedMember()
	if not lastJoinedMember then
		if table.getn(oldPartyMembers) < table.getn(q:GetPartyMembers()) then
			-- party member joined
			for id, name in ipairs(q:GetPartyMembers()) do
				if not q:TableContains(oldPartyMembers, name) then
					lastJoinedMember = name;
					return;
				end
			end
		elseif table.getn(oldRaidMembers[0]) < table.getn(q:GetRaidMembers()[0]) then
			-- raid member joined
			for id, name in ipairs(q:GetRaidMembers()[0]) do
				if not q:TableContains(oldRaidMembers[0], name) then
					lastJoinedMember = name;
					return;
				end
			end
		end
	end
	return lastJoinedMember;
end

function q:GetLastLeftMember()
	if not lastLeftMember then
		if table.getn(oldPartyMembers) > table.getn(q:GetPartyMembers()) then
			-- party member left
			for id, name in ipairs(oldPartyMembers) do
				if not q:TableContains(q:GetPartyMembers(), name) then
					lastLeftMember = name;
					return;
				end
			end
		elseif table.getn(oldRaidMembers[0]) > table.getn(q:GetRaidMembers()[0]) then
			-- raid member left
			for id, name in ipairs(oldRaidMembers[0]) do
				if not q:TableContains(q:GetRaidMembers()[0], name) then
					lastLeftMember = name;
					return;
				end
			end
		end
	end
	return lastLeftMember;
end

function q:GetBattleNetTag()
	if BNConnected() then
		local presenceID, toonID, currentBroadcast, bnetAFK, bnetDND = BNGetInfo();
		return toonID;
	end
end

function q:SendAddonMessage(addon, message, channel, target)
	addon = addon or q;
	C_ChatInfo.SendAddonMessage(addon.name, message, channel, target);
end

-- /run qualityoflife:DumpTable(qualityoflife:GetVisibleNamePlates())
function q:GetVisibleNamePlates()
	local namePlates = {}
	for n = 1, 80 do
		local namePlate = _G["NamePlate" .. n];
		if namePlate and namePlate:IsVisible() and namePlate.UnitFrame and UnitIsPlayer(namePlate.UnitFrame.unit) then
			namePlates[namePlate.UnitFrame.unit] = namePlate;
		end
	end
	return namePlates;
end

function q:GetPlayerNamePlate(playerName)
	for n = 1, 80 do
		local namePlate = _G["NamePlate" .. n];
		if namePlate and namePlate:IsVisible() and namePlate.UnitFrame and UnitIsPlayer(namePlate.UnitFrame.unit) then
			if UnitName(namePlate.UnitFrame.unit) == playerName then
				return namePlate;
			end
		end
	end
end